/*********************************************************/
/* file: sight.c                                         */
/*                                                       */
/* Copyright (C) 1994 Karl A Holland                     */
/*                                                       */
/* author: Karl A Holland                                */
/*                                                       */
/* created:   1 april 93                                 */
/*                                                       */
/* modified:  7 february 94                              */
/*                                                       */
/* contributing authors:                                 */
/*  if you add place your name here or below             */
/*                                                       */
/* description:                                          */
/*  sight.c contains subroutines for computing line of   */
/* sight information for a player. It is hoped that the  */
/* propagation of light via a stack is faster that       */
/* conventional methods of computing LOS, though the     */
/* results are not nearly as accurate                    */
/*                                                       */
/* LICENSING:                                            */
/*   a full license  is provided in the file LICENSE     */
/* which should have been included with this program when*/
/* it was copied. It may be found in the top directory   */
/* of any NORMAL distribution, named LICENSE             */
/*********************************************************/
#include <standard.h>

typedef struct coords 
{
  int             xl;
  int             yl;

}               coords;

/***************************************/
/* clear_sight                         */
/*   will set everything in the player */
/* structures line of sight information*/
/* to clear so that the next iteration */
/* of line of sight may fill it with   */
/* new information                     */
/***************************************/
void
clear_sight(playerstruct * pl)
{
  int             i,
                  j;
  for (i = 0; i < 7; i++)
    for (j = 0; j < 7; j++)
      pl->in_sight[i][j] = 1;
}

/***************************************************************/
/***** these defines are local to the set sight procedure ******/
/***************************************************************/
#define STACKSIZE 150

#define OUT_OF_SIGHT(x,y) ((x>rx) || (x< lx) || (y >ty) || (y <by))

#define PUSH(a,b,c,d) {xstack[++xtop]= a;ystack[++ytop]= b;tstack[++ttop]= c;dstack[++dtop]= d;} 

#define POP(x,y,t,d)  {x = xstack[xtop];xtop --;y = ystack[ytop];ytop--;t = tstack[ttop];ttop --; d = dstack[dtop]; dtop--;}

#define VIEW 1

#define VIEWL 3

#define VIEWR 5

#define BLOCKED 0

#define BLOCKEDL 2

#define BLOCKEDR 4

/***************************************************************/
/****************end set_sight local defines********************/
/***************************************************************/


void
set_sight(playerstruct * pl)
{
  int             cur_x,
                  cur_y,
                  type,
                  dir,
                  ldir,
                  rdir,
                  xtop = -1,
                  ytop = -1,
                  ttop = -1,
                  dtop = -1,
                  lx,
                  rx,
                  ty,
                  by,
                  px = pl->obj->x,
                  py = pl->obj->y,
                  ref = pl->obj->facing / 2,
                  xstack[STACKSIZE],
                  ystack[STACKSIZE],
                  tstack[STACKSIZE],
                  dstack[STACKSIZE];
  map            *pmap = pl->obj->mapin;

  clear_sight(pl);
  
  lx = px + (tvecs[ref][0] * 5) - (tvecs[ref][1] * 3);
  rx = px - (tvecs[ref][0] * 1) + (tvecs[ref][1] * 3);
  ty = py + (tvecs[ref][1] * 5) + (tvecs[ref][0] * 3);
  by = py - (tvecs[ref][1] * 1) - (tvecs[ref][0] * 3);
  
  if (rx < lx)
    {
      cur_x = lx;
      lx = rx;
      rx = cur_x;
    }
  
  rx = (rx >= pmap->boundx ? pmap->boundx - 1 : rx);
  lx = (lx < 0 ? 0 : lx);
  
  pl->lowx = lx;
  pl->hix= rx;
  
  if (ty < by)
    {
      cur_y = by;
      by = ty;
      ty = cur_y;
    }
  ty =(ty >= pmap->boundy ? pmap->boundy - 1 : ty);
  by =(by < 0 ? 0 : by);
  pl->lowy =by;
  pl->hiy = by;
  for (dir = 0; dir < 8; dir+=2) 
    { 
      ldir = (dir + 7) % 8; 
      PUSH(px + xdeltas[dir], py + ydeltas[dir], VIEW, dir);
      PUSH(px + xdeltas[ldir], py + ydeltas[ldir], VIEWL, dir); 
    }
  
  
  
  while (xtop + 1)
    {
      POP(cur_x, cur_y, type, dir);
      if (!OUT_OF_SIGHT(cur_x, cur_y) && pl->in_sight[cur_x - lx][cur_y - by])
	{
	  rdir = (dir + 1) % 8;
	  ldir = (dir + 7) % 8;
	  switch (type)
	    {
	    case (VIEW):
	      if (MAP_VIEW_BLOCKED(pmap, cur_x, cur_y, 1) 
		  || MAP_VIEW_BLOCKED(pmap, cur_x, cur_y, 2))
		{
		  PUSH(cur_x + xdeltas[rdir], cur_y + ydeltas[rdir],
		       BLOCKEDR, dir);
		  PUSH(cur_x + xdeltas[ldir], cur_y + ydeltas[ldir],
		       BLOCKEDL, dir);
		  PUSH(cur_x + xdeltas[dir], cur_y + ydeltas[dir],
		       BLOCKED, dir);
		  if (((pl->obj->facing+6)% 8) == 
		      dir || ((pl->obj->facing+4)% 8) == dir)
		    pl->in_sight[cur_x - lx][cur_y - by] = 2;
		} 
	      else
		{
		  PUSH(cur_x + xdeltas[rdir], cur_y + ydeltas[rdir],
		       VIEWR, dir);
		  PUSH(cur_x + xdeltas[ldir], cur_y + ydeltas[ldir],
		       VIEWL, dir);
		  PUSH(cur_x + xdeltas[dir], cur_y + ydeltas[dir],
		       VIEW, dir);
		}
	      break;
	    case (VIEWR):
	      if (MAP_VIEW_BLOCKED(pmap, cur_x, cur_y, 1) 
		  || MAP_VIEW_BLOCKED(pmap, cur_x, cur_y, 2))
		{
		  PUSH(cur_x + xdeltas[rdir], cur_y + ydeltas[rdir],
		       BLOCKEDR, dir);
		  if (((pl->obj->facing+4)% 8)== dir)
		    pl->in_sight[cur_x - lx][cur_y - by] = 2;
		}
	      else
		{
		  PUSH(cur_x + xdeltas[rdir], cur_y + ydeltas[rdir],
		       VIEWR, dir);
		}
	      break;
	    case (VIEWL):
	      if (MAP_VIEW_BLOCKED(pmap, cur_x, cur_y, 1) 
		  ||MAP_VIEW_BLOCKED(pmap, cur_x, cur_y, 2))
		{
		  PUSH(cur_x + xdeltas[ldir], cur_y + ydeltas[ldir],
		       BLOCKEDL, dir);
		  if (((pl->obj->facing+6)% 8) == dir 
		      || ((pl->obj->facing+4)% 8) == dir)
		    pl->in_sight[cur_x - lx][cur_y - by] = 2;
		} 
	      else
		{
		  PUSH(cur_x + xdeltas[ldir], cur_y + ydeltas[ldir],
		       VIEWL, dir);
		}
	      break;
	    case (BLOCKED):
	      pl->in_sight[cur_x - lx][cur_y - by] = 0;
	      PUSH(cur_x + xdeltas[rdir], cur_y + ydeltas[rdir],
		   BLOCKEDR, dir);
	      PUSH(cur_x + xdeltas[ldir], cur_y + ydeltas[ldir], 
		   BLOCKEDL, dir);
	      PUSH(cur_x + xdeltas[dir], cur_y + ydeltas[dir],
		   BLOCKED, dir);
	      break;
	    case (BLOCKEDL):
	      pl->in_sight[cur_x - lx][cur_y - by] = 0;
	      PUSH(cur_x + xdeltas[ldir], cur_y + ydeltas[ldir],
		   BLOCKEDL, dir);
	      break;
	    case (BLOCKEDR):
	      pl->in_sight[cur_x - lx][cur_y - by] = 0;
	      PUSH(cur_x + xdeltas[rdir], cur_y + ydeltas[rdir],
		   BLOCKEDR, dir);
	      break;
	    }
	}
    }
}
